require 'rails_helper'
require 'faker'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  VALID_OPERATORS = %w[
    $in
    $nin
    $ne
    $gt
    $gte
    $lt
    $lte
    $regex
  ].freeze
  
  ACCEPTABLE_PARAMS = %w[
    ids
    department
    creation_date
    unique_number
    is_active
  ].freeze
  
  REGEX_PARAMS = %w[
  ].freeze
  
  MUTATE_PARAMS = {
      department: 'String',
      creation_date: 'DateTime',
      unique_number: 'String'
  }.freeze
  
  UNIQUES_NUMBERS = ["0130418317255AL004", "0281118062762AL004", "0281217055366AL004", "0330218247547GDPUB", "0330617221751AL004", "0541018104918AL004", "0590519548544GDPUB", "0600116076461GDPUB",
                     "0270118080985AL004", "0281118062911AL004", "0300418091847AL004", "0330319288338AL004", "0330718264892GDPUB", "0541117094439AL004", "0591018505670AL004",
                     "0281117054754GDPUB", "0281217055070AL004", "0330117206320GDPUB", "0330519295592ACLOG", "0330817227566GDPUB", "0590318466943AL004", "0591118512641GDPUB"]
  
  RSpec.describe Api::Dmo::NudsController, :type => :api do
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper
    
    let!(:responsible) {create(:user, roles: ['Administrator'])}
    let!(:user) {create(:user)}
    let!(:user1) {create(:user, auth_roles: ['Demandeur'])}
    #TESTS BASIQUES D'API
    
    describe 'POST /nuds' do
      describe 'Create nud with minimal valid datas' do
        let!(:applicant) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant)}
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
              data: {
                  attributes: {
                      user_id: responsible.id.to_s,
                      als_id: request1.als_id.to_s
                  }
              }
          }
        end
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end
      end
      
      describe 'Create with alreay existing pair of uniq_nud and hosuing_request_id' do
        unique_number = Faker::IDNumber
        let!(:applicant) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant)}
        let!(:nud_test) {create(:nud, housing_request: request1, unique_number: unique_number)}
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
              data: {
                  attributes: {
                      user_id: responsible.id.to_s,
                      als_id: request1.als_id.to_s,
                      unique_number: unique_number,
                      creation_date: '1867-03-24'
                  }
              }
          }
        end
        
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
        end
        
        it 'doesn\'t create another nud' do
          expect(Hermes::Nud.count).to be 1
        end
        
        it 'updates nud_test' do
          data = json['data']
          expect(data['id']).to eq nud_test.id.to_s
          expect(Date.parse(data['attributes']['creation_date'])).to eq(Date.parse('1867-03-24'))
        end
      end
      
      MUTATE_PARAMS.each do |param, type|
        describe 'post nud with ' + param.to_s + ' (' + type.to_s + ') is valid' do
          let!(:applicant) {create(:applicant)}
          let!(:request1) {create(:housing_request, main_applicant: applicant)}
          let(:params) do
            attr = {}
            attr[param] = get_typed_value(type)
            attr
          end
          before :each do
            params['user_id'] = user.id.to_s
            params['als_id'] = request1.als_id.to_s
            authenticated_header(request, responsible)
            post 'create', {
                data: {
                    attributes: params
                }
            }
          end
          it 'returns status code 201 and the value appears in json' do
            expect(response).to have_http_status(201)
            if (type == 'Date' || type == 'DateTime')
              expect(json['data']['attributes'][param.to_s]).not_to be_empty
              expect(json['data']['attributes'][param.to_s]).not_to eq('')
            else
              expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type))
            end
          end
        end
      end
      #housing_request_id:'BSON::ObjectId',
      describe 'post nud with request' do
        let!(:applicant1) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant1)}
        before :each do
          authenticated_header(request, responsible)
          post 'create', {
              data: {
                  attributes: {
                      user_id: responsible.id.to_s,
                      als_id: request1.als_id.to_s,
                  }
              }
          }
        end
        it 'returns status code 201 and the value appears in json' do
          expect(response).to have_http_status(201)
          expect(json['data']['attributes']['housing_request_id']).to eq(request1.id.to_s)
        end
      end
    end
    
    describe 'GET /lessors' do
      describe 'GET /lessors' do
        let!(:applicant1) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant1)}
        let!(:nud1) {create(:nud, user_id: responsible.id.to_s, housing_request: request1)}
        let!(:nud3) {create(:nud, user_id: responsible.id.to_s, housing_request: request1)}
        let!(:nud2) {create(:nud, user_id: responsible.id.to_s, housing_request: request1)}
        before :each do
          authenticated_header(request, responsible)
          get 'index'
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
          expect(json['data'].length).to eq(3)
          expect(json['meta']['pagination']).to be_present
          expect(json['meta']['pagination']['total_objects']).to eq(3)
        end
      end
      
      describe 'GET /nuds' do
        let!(:applicant1) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant1)}
        let!(:nud1) {create(:nud, user_id: responsible.id.to_s, housing_request: request1)}
        let!(:nud3) {create(:nud, user_id: responsible.id.to_s, housing_request: request1)}
        let!(:nud2) {create(:nud, user_id: responsible.id.to_s, housing_request: request1)}
        before :each do
          authenticated_header(request, responsible)
          get 'index', {ids: [nud1.id.to_s, nud2.id.to_s]}
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
          expect(json['data'].length).to eq(2)
        end
      end
      
      ACCEPTABLE_PARAMS.each do |acceptable_param|
        type = MUTATE_PARAMS[acceptable_param.to_sym] || MUTATE_PARAMS[acceptable_param.to_s]
        if type
          describe 'GET /nuds with sort on field ' + acceptable_param + ' works ' do
            let!(:applicant1) {create(:applicant)}
            let!(:request1) {create(:housing_request, main_applicant: applicant1)}
            let!(:nud3) {create(:nud, acceptable_param => get_typed_value(type, 2), user_id: responsible.id.to_s, housing_request: request1)}
            let!(:nud1) {create(:nud, acceptable_param => get_typed_value(type, 0), user_id: responsible.id.to_s, housing_request: request1)}
            let!(:nud2) {create(:nud, acceptable_param => get_typed_value(type, 1), user_id: responsible.id.to_s, housing_request: request1)}
            before :each do
              authenticated_header(request, responsible)
              get 'index', {sort: {acceptable_param => 1}}
            end
            it 'returns sorted results' do
              expect(response.status).to eq(200)
              if type == 'DateTime'
                nud_attribute = json['data'].map {|unit| DateTime.parse(unit['attributes'][acceptable_param])}
                expect(nud_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              elsif type == 'Date'
                nud_attribute = json['data'].map {|unit| Date.parse(unit['attributes'][acceptable_param])}
                expect(nud_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              else
                nud_attribute = json['data'].map {|unit| unit['attributes'][acceptable_param]}
                expect(nud_attribute).to eq([get_typed_value(type, 0), get_typed_value(type, 1), get_typed_value(type, 2)])
              end
            end
          end
        end
      end
      describe 'GET /nuds with operators' do
        let!(:applicant1) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant1)}
        let!(:nud1) {create(:nud, department: '24', user_id: responsible.id.to_s, housing_request: request1)}
        let!(:nud3) {create(:nud, department: '2A', user_id: responsible.id.to_s, housing_request: request1)}
        let!(:nud2) {create(:nud, department: '65', user_id: responsible.id.to_s, housing_request: request1)}
        
        describe 'with $in' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {department: {'$in': ['24', '73']}}
          end
          it 'returns results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
        describe 'with $ne' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {department: {'$ne': '2A'}}
          end
          it 'returns sorted results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(2)
          end
        end
        describe 'with $regex' do
          before :each do
            authenticated_header(request, responsible)
            get 'index', {department: {'$regex': /^6/}}
          end
          it 'returns sorted results' do
            expect(response.status).to eq(200)
            expect(json['meta']['pagination']['total_objects']).to eq(1)
          end
        end
      end
    end
    
    describe 'GET /nuds/:id' do
      describe 'verif it work' do
        let!(:applicant1) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant1)}
        let!(:nud1) {create(:nud, department: '24', user_id: responsible.id.to_s, housing_request: request1)}
        before :each do
          authenticated_header(request, responsible)
          get 'show', {id: nud1.id}
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
        end
        it 'returns a valid json' do
          expect(json).not_to be_empty
        end
      end
      describe 'verif it work not work with a non authorized user' do
        let!(:applicant1) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant1)}
        let!(:nud1) {create(:nud, department: '24', user_id: responsible.id.to_s, housing_request: request1)}
        
        before :each do
          authenticated_header(request, user)
          get 'show', {id: nud1.id}
        end
        it 'returns status code 404' do
          expect(response).to have_http_status(404)
        end
      end
    end
    
    describe 'PATCH /nuds/:id' do
      nud_params = {}
      MUTATE_PARAMS.each do |param, type|
        nud_params[param] = get_typed_value(type)
      end
      
      let!(:applicant1) {create(:applicant)}
      let!(:request1) {create(:housing_request, main_applicant: applicant1)}
      let!(:nud) {create(:nud, **nud_params, user_id: responsible.id.to_s, housing_request: request1)}
      MUTATE_PARAMS.each do |param, type|
        describe 'patch request with ' + param.to_s + ' ' + type.to_s + 'is valid' do
          let(:params) do
            attr = {}
            attr[param] = get_typed_value(type, 2)
            attr
          end
          before :each do
            authenticated_header(request, responsible)
            patch 'update', {
                id: nud.id.to_s,
                data: {
                    attributes: params
                }
            }
          end
          it 'returns status code 200 and the value appears in json' do
            expect(response).to have_http_status(200)
            if (type == 'Date' || type == 'DateTime')
              
              expect(json['data']['attributes'][param.to_s]).not_to be_empty
              expect(json['data']['attributes'][param.to_s]).not_to eq('')
            else
              expect(json['data']['attributes'][param.to_s]).to eq(get_typed_value(type, 2))
            end
          end
        end
      end
      
      describe 'DELETE /nuds/:id' do
        let!(:applicant1) {create(:applicant)}
        let!(:request1) {create(:housing_request, main_applicant: applicant1)}
        let!(:nud) {create(:nud, **nud_params, user_id: responsible.id.to_s, housing_request: request1)}
        before :each do
          authenticated_header(request, responsible)
          
          delete 'destroy', {
              id: nud.id.to_s
          }
        end
        it 'returns status code 204' do
          expect(response).to have_http_status(204)
        end
      end
      
        describe "with unique_number" do
        UNIQUES_NUMBERS = ["0300418091847AL004"]
          UNIQUES_NUMBERS.each do |unique_number|
          describe " = " + unique_number + ": " + "POST api/dmo/nuds/:unique_number/send_pin_code_on_mail" do
            let!(:hri) {create(:housing_request_initializer, unique_number: unique_number)}
    
            before :each do
              authenticated_header(request, user)
              post 'send_pin_code_on_mail', {
                  id: hri.unique_number,
              }
            end
            it 'returns status code 200' do
              expect(response).to have_http_status(200)
              hri.reload
              expect(hri.hashed_pin_code).not_to be_empty
              mailer_jobs = enqueued_jobs.select {|job| job[:job] == ActionMailer::DeliveryJob}
              expect(mailer_jobs.length).to be >= 1
              mail = perform_enqueued_jobs {ActionMailer::DeliveryJob.perform_now(*mailer_jobs.first[:args])}
              expect(mail).not_to eq(nil)
            end
          end
          describe " = " + unique_number + ": " +'POST api/dmo/nuds/:unique_number/authentify_with_pin_code' do
            let!(:hri) {create(:housing_request_initializer, unique_number: unique_number)}
            let!(:pincode) {hri.send_pin_code_on_mail}
    
            before :each do
              authenticated_header(request, user)
      
              post 'authentify_with_pin_code', {
                  id: hri.reload.unique_number,
                  pin_code: pincode
              }
            end
            it 'returns status code 200' do
              expect(response).to have_http_status(200)
              ap json
              expect(json).not_to be_empty
              
              expect(json["data"]["type"]).to eq("housing_requests")
              #expect(json["data"]["attributes"]["file_upload_from_sne_complete"]).to eq(true)
            end
          end
        end
      end
    end
  end
  
  RSpec.describe Api::Dmo::NudsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
    require_relative '../../swagger_helper'
    TAGS_API_DMO_NUDS = 'API_DMO_NUDS'.freeze
    include JwtAuthHelper
    extend HermesSwaggerJsonParams
    include SwaggerJsonAttributes
    let!(:responsible) {create(:user, first_name: 'Martin', last_name: 'Durand', email: 'martin.durand@gmail.com', roles: ['Administrator'])}
    let!(:user) {create(:user, first_name: 'Jackie', last_name: 'Mike', email: 'jackie.mike@gmail.com')}
    
    context 'as admin' do
      let!(:admin) {create(:administrator)}
      let!(:applicant1) {create(:applicant, email: "bibi@laposte.net")}
      let!(:request1) {create(:housing_request, main_applicant: applicant1)}
      
      path '/api/dmo/nuds' do
        post 'Create nud' do
          tags TAGS_API_DMO_NUDS
          produces 'application/json'
          consumes 'application/json'
          security [Bearer: {}]
          json_params = {}
          json_params[:user_id] = {type: :string}
          json_params[:housing_request_id] = {type: :string}
          json_params[:als_id] = {type: :string}
          json_params[:creation_date] = {type: :string}
          json_params[:unique_number]
          MUTATE_PARAMS.except(:creation_date).each do |key, val|
            json_params[key] = {type: nice_type(val)}
          end
          parameter(
              name: :params,
              in: :body,
              schema: jsonify_params(json_params),
              description: '<span>' \
            '<div class="opblock opblock-delete" style="padding: 4px; align-items: center; display: flex">' \
            '<span class="opblock-summary-method" style="margin-right: 4px">WARNING</span>' \
            '<span class="no-style">Creating a nud that contains an already existing pair of `uniq_number` <==> ' \
              '`housing_request_id`, will find and update the item with this already existing pair' +
                  ' instead of creating a new item - avoid duplicates.</span></div></span>'
          )
          
          response '201', 'nud created' do
            let(:Authorization) {"Bearer #{token_for(responsible)}"}
            let(:params) do
              jsonapi_attributes(
                  housing_request_id: request1.id.to_s
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'
            
            after do |_example|
              json = JSON.parse(response.body)
              
              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'nuds'
            end
          end
        end
      end
      
      path '/api/dmo/nuds' do
        get 'Retrieves all nuds' do
          tags TAGS_API_DMO_NUDS
          produces 'application/json'
          
          parameter name: 'ids[]', in: :query, type: :array, items: {type: :string}, collectionFormat: :multi, required: false
          ACCEPTABLE_PARAMS.each do |acceptable_param|
            type = ''
            if MUTATE_PARAMS.include?(acceptable_param)
              type = nice_type(MUTATE_PARAMS[acceptable_param])
            end
            parameter name: acceptable_param.to_s, in: :query, type: type, required: false
          end
          
          # VALID_OPERATORS.each do |valid_operator|
          #   parameter( name: valid_operator.to_s, in: :query,type: :object, required: false)
          # end
          
          security [Bearer: {}]
          paginator
          
          response '200', 'nuds found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let!(:nuds) {create_list(:nud, 5, housing_request: request1)}
            
            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) {nuds}
            end
          end
        end
      end
      
      path '/api/dmo/nuds/{id}' do
        get 'Retrieves a single nud by id' do
          tags TAGS_API_DMO_NUDS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          
          response '200', 'nud found' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:nud) {create(:nud, housing_request: request1)}
            let(:id) {nud.id}
            
            include_context 'with integration test'
            include_context 'valid jsonapi ressource' do
              let(:expected_ressource) {nud}
            end
          end
        end
      end
      
      path '/api/dmo/nuds/{id}' do
        put 'edit nud/{id}' do
          tags TAGS_API_DMO_NUDS
          produces 'application/json'
          consumes 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          json_params = {}
          json_params[:housing_request_id] = {type: :string}
          json_params[:is_active] = {type: :boolean, default: true}
          MUTATE_PARAMS.each do |key, val|
            json_params[key] = {type: nice_type(val)}
          end
          
          parameter name: :params, in: :body, schema: jsonify_params(json_params)
          
          response '200', 'nud created' do
            let(:Authorization) {"Bearer #{token_for(responsible)}"}
            let(:nud) {create(:nud, housing_request: request1)}
            let(:id) {nud.id}
            let(:params) do
              jsonapi_attributes(
                  id: id
              )
            end
            include_context 'with integration test'
            include_context 'valid jsonapi'
            include_context 'not containings errors'
            after do |_example|
              json = JSON.parse(response.body)
              
              expect(json['data']).to be_a Hash
              expect(json['data']['id']).to_not be nil
              expect(json['data']['type']).to eq 'nuds'
            end
          end
        end
      end
      
      path '/api/dmo/nuds/{id}' do
        delete 'Delete a single nud by id' do
          tags TAGS_API_DMO_NUDS
          produces 'application/json'
          parameter name: :id, in: :path, type: :string
          security [Bearer: {}]
          
          response '204', 'nud deleted' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:id) {create(:nud, responsible: admin.id, housing_request: request1).id}
            run_test!
          end
        end
      end
      path '/api/dmo/nuds/{numero_unique_departemental}/send_pin_code_on_mail' do
        post 'send_pin_code_on_mail' do
          tags TAGS_API_DMO_NUDS
          produces 'application/json'
          parameter name: :numero_unique_departemental, in: :path, type: :string, required: true
          security [Bearer: {}]
          
          response '200', 'pin code send on email' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:numero_unique_departemental) {UNIQUES_NUMBERS[0]}
            
            run_test!
          
          end
=begin
          response '400', 'not found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let(:nud) { create(:nud, housing_request: request1, unique_number: "3REFLIDKFF") }
            let(:numero_unique_departemental) { "ZZZ"}

            run_test!
          end
=end
        end
      end
      path '/api/dmo/nuds/{numero_unique_departemental}/authentify_with_pin_code' do
        post 'send_pin_code_on_mail' do
          tags TAGS_API_DMO_NUDS
          produces 'application/json'
          parameter name: :numero_unique_departemental, in: :path, type: :string, required: true
          security [Bearer: {}]
          parameter name: :pin_code, in: :body, type: :string, requied: true

=begin
          response '200', 'pin code validated' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let(:nud) { create(:nud, housing_request: request1, unique_number: "3REFLIDKFF") }
            let(:numero_unique_departemental) {  nud.unique_number}
            let(:pin_code) {nud.send_pin_code_on_mail}

            run_test!

          end
=end
          response '422', 'unable to validate pin_code' do
            let(:Authorization) {"Bearer #{token_for(admin)}"}
            let(:nud) {create(:nud, housing_request: request1, unique_number: "3REFLIDKFF")}
            let(:numero_unique_departemental) {nud.unique_number}
            let(:pin_code) {"slip"}
            run_test!
          end
        end
      end
    end
  end
end
